import re
import types

def add_getlist(obj):
    #if not hasattr(obj, "getlist"):
    #    def _getlist(key, default=None):
    #        return [obj.get(key, default)]
#        setattr(obj, "getlist", types.MethodType(_getlist, obj))
    return obj

_extract_streetnr = re.compile(r"([0-9]+)\s*$")
def split_streetnr(address, fallback=None):
    ret = _extract_streetnr.search(address[-15:])
    if ret:
        return address[:(ret.start()-15)].strip(), ret.group(0)
    else:
        return address.strip(), fallback

def getter_prefixed_address(prefix):
    """ create getter for prefixed address format """
    first_name = "{}_first_name".format(prefix)
    last_name = "{}_last_name".format(prefix)
    address_1 = "{}_address_1".format(prefix)
    address_2 = "{}_address_2".format(prefix)
    city = "{}_city".format(prefix)
    postcode = "{}_postcode".format(prefix)
    country_code = "{}_country_code".format(prefix)
    country_area = "{}_country_area".format(prefix)
    def _get_address(self):
        return {
            "first_name": getattr(self, first_name, None),
            "last_name": getattr(self, last_name, None),
            "address_1": getattr(self, address_1, None),
            "address_2": getattr(self, address_2, None),
            "city": getattr(self, city, None),
            "postcode": getattr(self, postcode, None),
            "country_code": getattr(self, country_code, None),
            "country_area": getattr(self, country_area, None)}
    return _get_address


CARD_TYPES = [
    (r'^4[0-9]{12}(?:[0-9]{3})?$', 'visa', 'VISA'),
    (r'^5[1-5][0-9]{14}$', 'mastercard', 'MasterCard'),
    (r'^6(?:011|5[0-9]{2})[0-9]{12}$', 'discover', 'Discover'),
    (r'^3[47][0-9]{13}$', 'amex', 'American Express'),
    (r'^(?:(?:2131|1800|35\d{3})\d{11})$', 'jcb', 'JCB'),
    (r'^(?:3(?:0[0-5]|[68][0-9])[0-9]{11})$', 'diners', 'Diners Club'),
    (r'^(?:5[0678]\d\d|6304|6390|67\d\d)\d{8,15}$', 'maestro', 'Maestro')]


def get_credit_card_issuer(number):
    for regexp, card_type, name in CARD_TYPES:
        if re.match(regexp, number):
            return card_type, name
    return None, None



from unittest import TestCase
from unittest.mock import patch, NonCallableMock

from . import get_base_url

class TestHelpers(TestCase):
    @patch('web_payments.django.PAYMENT_USES_SSL', new_callable=NonCallableMock)
    @patch('web_payments.django.PAYMENT_HOST', new_callable=NonCallableMock)
    def test_text_get_base_url(self, host, use_ssl):
        host.__str__ = lambda x: "example.com/string"
        use_ssl = False
        self.assertEqual(get_base_url(), "https://example.com/string")

    @patch('web_payments.django.PAYMENT_USES_SSL', new_callable=NonCallableMock)
    @patch('web_payments.django.PAYMENT_HOST')
    def test_callable_get_base_url(self, host, use_ssl):
        host.return_value = "example.com/callable"
        use_ssl = False
        self.assertEqual(get_base_url(), "https://example.com/callable")



import re
from . import NotInitialized

is_initialized = False

PROVIDER_CACHE = {}

PAYMENT_VARIANTS_API = {
    'default': ('web_payments_dummy.DummyProvider', {})}

def get_language():
    raise NotInitialized()

def get_base_url(variant=None):
    raise NotInitialized()

def get_payment_model():
    '''
    Return the Payment model of default backend
    '''
    raise NotInitialized()

def provider_factory(variant):
    '''
    Return the provider instance based on variant
    '''
    handler, config = PAYMENT_VARIANTS_API.get(variant, (None, None))
    if not handler:
        raise ValueError('Payment variant does not exist: %s' %
                         (variant,))
    if variant not in PROVIDER_CACHE:
        module_path, class_name = handler.rsplit('.', 1)
        module = __import__(
            str(module_path), globals(), locals(), [str(class_name)])
        class_ = getattr(module, class_name)
        PROVIDER_CACHE[variant] = class_(**config)
    return PROVIDER_CACHE[variant]



from decimal import Decimal
from unittest import TestCase
from unittest.mock import patch, NonCallableMock

from . import core
from .status import PaymentStatus
from .forms import CreditCardPaymentFormWithName, PaymentForm
from .testcommon import create_test_payment

BasePayment = create_test_payment()

class TestProviderFactory(TestCase):

    def test_provider_factory(self):
        core.provider_factory('default')

    def test_provider_does_not_exist(self):
        self.assertRaises(ValueError, core.provider_factory, 'fake_provider')


class TestBasePayment(TestCase):

    def test_payment_attributes(self):
        payment = BasePayment(
            extra_data='{"attr1": "test1", "attr2": "test2"}')
        self.assertEqual(payment.attrs.attr1, "test1")
        self.assertEqual(payment.attrs.attr2, 'test2')
        self.assertEqual(getattr(payment.attrs, "attr5", None), None)
        self.assertEqual(hasattr(payment.attrs, "attr7"), False)

    def test_capture_with_wrong_status(self):
        payment = BasePayment(variant='default', status=PaymentStatus.WAITING)
        self.assertRaises(ValueError, payment.capture)

    @patch('web_payments_dummy.DummyProvider.capture')
    def test_capture_preauth_successfully(self, mocked_capture_method):
        amount = Decimal('20')
        with patch.object(BasePayment, 'save') as mocked_save_method:
            mocked_save_method.return_value = None
            mocked_capture_method.return_value = amount

            captured_amount = Decimal('0')
            payment = BasePayment(variant='default', captured_amount=captured_amount,
                                  status=PaymentStatus.PREAUTH)
            payment.capture(amount)

            self.assertEqual(payment.status, PaymentStatus.CONFIRMED)
            self.assertEqual(payment.captured_amount, amount)
        self.assertEqual(mocked_capture_method.call_count, 1)

    @patch('web_payments_dummy.DummyProvider.capture')
    def test_capture_preauth_without_amount(self, mocked_capture_method):
        amount = None
        with patch.object(BasePayment, 'save') as mocked_save_method:
            mocked_save_method.return_value = None
            mocked_capture_method.return_value = amount

            captured_amount = Decimal('0')
            status = PaymentStatus.PREAUTH
            payment = BasePayment(variant='default', status=status,
                                  captured_amount=captured_amount)
            payment.capture(amount)

            self.assertEqual(payment.status, status)
            self.assertEqual(payment.captured_amount, captured_amount)
        self.assertEqual(mocked_capture_method.call_count, 1)

    def test_release_with_wrong_status(self):
        payment = BasePayment(variant='default', status=PaymentStatus.WAITING)
        self.assertRaises(ValueError, payment.release)

    @patch('web_payments_dummy.DummyProvider.release')
    def test_release_preauth_successfully(self, mocked_release_method):
        with patch.object(BasePayment, 'save') as mocked_save_method:
            mocked_save_method.return_value = None

            payment = BasePayment(variant='default', status=PaymentStatus.PREAUTH)
            payment.release()
            self.assertEqual(payment.status, PaymentStatus.REFUNDED)
        self.assertEqual(mocked_release_method.call_count, 1)

    def test_refund_with_wrong_status(self):
        payment = BasePayment(variant='default', status=PaymentStatus.WAITING)
        self.assertRaises(ValueError, payment.refund)

    def test_refund_too_high_amount(self):
        payment = BasePayment(variant='default', status=PaymentStatus.CONFIRMED,
                              captured_amount=Decimal('100'))
        self.assertRaises(ValueError, payment.refund, Decimal('200'))

    @patch('web_payments_dummy.DummyProvider.refund')
    def test_refund_without_amount(self, mocked_refund_method):
        refund_amount = None
        with patch.object(BasePayment, 'save') as mocked_save_method:
            mocked_save_method.return_value = None
            mocked_refund_method.return_value = refund_amount

            captured_amount = Decimal('200')
            status = PaymentStatus.CONFIRMED
            payment = BasePayment(variant='default', status=status,
                                  captured_amount=captured_amount)
            payment.refund(refund_amount)
            self.assertEqual(payment.status, status)
            self.assertEqual(payment.captured_amount, captured_amount)
        self.assertEqual(mocked_refund_method.call_count, 1)

    @patch('web_payments_dummy.DummyProvider.refund')
    def test_refund_partial_success(self, mocked_refund_method):
        refund_amount = Decimal('100')
        captured_amount = Decimal('200')
        status = PaymentStatus.CONFIRMED
        with patch.object(BasePayment, 'save') as mocked_save_method:
            mocked_save_method.return_value = None
            mocked_refund_method.return_value = refund_amount

            payment = BasePayment(variant='default', status=status,
                                  captured_amount=captured_amount)
            payment.refund(refund_amount)
            self.assertEqual(payment.status, status)
            self.assertEqual(payment.captured_amount, Decimal('100'))
        self.assertEqual(mocked_refund_method.call_count, 1)

    @patch('web_payments_dummy.DummyProvider.refund')
    def test_refund_fully_success(self, mocked_refund_method):
        refund_amount = Decimal('200')
        captured_amount = Decimal('200')
        with patch.object(BasePayment, 'save') as mocked_save_method:
            mocked_save_method.return_value = None
            mocked_refund_method.return_value = refund_amount

            payment = BasePayment(variant='default', status=PaymentStatus.CONFIRMED,
                                  captured_amount=captured_amount)
            payment.refund(refund_amount)
            self.assertEqual(payment.status, PaymentStatus.REFUNDED)
            self.assertEqual(payment.captured_amount, Decimal('0'))
        self.assertEqual(mocked_refund_method.call_count, 1)


class TestCreditCardPaymentForm(TestCase):

    def setUp(self):
        self.credit_data = {
                'name': 'John Doe',
                'number': '4716124728800975',
                'expiration': '2020-05',
                'cvv2': '123'}

    def test_form_verifies_card_number(self):
        form = CreditCardPaymentFormWithName(data=self.credit_data)
        self.assertTrue(form.validate())

    def test_form_raises_error_for_invalid_card_number(self):
        data = dict(self.credit_data)
        data.update({'number': '1112223334445556'})
        form = CreditCardPaymentFormWithName(data=data)
        self.assertFalse(form.validate())
        self.assertIn('number', form.errors)

    def test_form_raises_error_for_invalid_cvv2(self):
        data = dict(self.credit_data)
        data.update({'cvv2': '12345'})
        form = CreditCardPaymentFormWithName(data=data)
        self.assertFalse(form.validate())
        self.assertIn('cvv2', form.errors)

"""
class TestPaymentForm(TestCase):

    def test_form_contains_hidden_fields(self):
        data = {
            'field1': 'value1',
            'field2': 'value2',
            'field3': 'value3',
            'field4': 'value4'}

        form = PaymentForm(data=data, hidden_inputs=True)
        self.assertEqual(len(form.fields), len(data))
        self.assertEqual(form.fields['field1'].initial, 'value1')
"""




from datetime import datetime
from decimal import Decimal
from .logic import BasePaymentLogic
from .status import PaymentStatus
from . import PurchasedItem
from .utils import getter_prefixed_address
from unittest.mock import Mock

def create_test_payment(**_kwargs):
    class TestPayment(Mock, BasePaymentLogic):
        id = 523
        pk = id
        description = 'payment'
        currency = 'USD'
        delivery = Decimal(10.8)
        status = PaymentStatus.WAITING
        message = ""
        tax = Decimal(10)
        token = "354338723"
        total = Decimal(100)
        captured_amount = Decimal("0.0")
        extra_data = ""
        variant = "undefined"
        transaction_id = None
        created = datetime.now()
        modified = datetime.now()

        billing_first_name = 'John'
        billing_last_name = 'Smith'
        billing_address_1 = 'JohnStreet 23'
        billing_address_2 = ''
        billing_city = 'Neches'
        billing_postcode = "75779"
        billing_country_code = "US"
        billing_country_area = "Tennessee"
        billing_email = "example@example.com"
        customer_ip_address = "192.78.6.6"

        get_billing_address = getter_prefixed_address("billing")
        get_shipping_address = get_billing_address

        def get_purchased_items(self):
            return [
                PurchasedItem(
                    name='foo', quantity=10, price=Decimal('20'),
                    currency='USD', sku='bar')]

        def get_failure_url(self):
            return 'http://cancel.com'

        def get_process_url(self):
            return 'http://example.com'

        def get_success_url(self):
            return 'http://success.com'

        def save(self):
            pass
    # workaround limitation in python
    for key, val in _kwargs.items():
        setattr(TestPayment, key, val)
    return TestPayment



from django.apps import AppConfig
from .. import core

class WebPaymentsConfig(AppConfig):
    name = 'web_payments'
    def ready(self):
        if not core.is_initialized:
            from . import load_settings
            load_settings()



from collections import namedtuple

PurchasedItem = namedtuple('PurchasedItem',
                           ['name', 'quantity', 'price', 'currency', 'sku'])

HttpRequest = namedtuple('HttpRequest',
                         ['method', 'GET', 'POST', 'content_type'])

default_app_config = 'web_payments.django.apps.WebPaymentsConfig'


class RedirectNeeded(Exception):
    pass

class ExternalPostNeeded(Exception):
    pass

class PaymentError(Exception):
    def __init__(self, message, code=None, gateway_message=None):
        super().__init__(message)
        self.code = code
        self.gateway_message = gateway_message

class NotInitialized(NotImplementedError):
    pass

class NotSupported(Exception):
    pass




import html
from wtforms import widgets, Field


class PlainTextWidget(object):
    """
    """
    def __call__(self, field, value=None, **kwargs):
        if not value:
            value = field._value()
        return widgets.HTMLString('<div {}>{}</div>'.format(widgets.html_params(**kwargs), html.escape(value)))

class TextField(Field):
    widget = PlainTextWidget()





from wtforms import StringField, ValidationError
from web_payments.fields import TextField
from web_payments.forms import PaymentForm
from .translation import translation
_ = translation.gettext_lazy

class OrderIdForm(PaymentForm):
    # only shown, return is ignored
    order = TextField(label=_("Please supply as reference"))


class IBANBankingForm(PaymentForm):
    # only shown, return is ignored
    iban = TextField(label="IBAN")
    bic = TextField(label="BIC")
    order = TextField(label=_("Please supply as reference"))




import datetime

from wtforms import Form, validators, ValidationError
from wtforms import StringField, DateField

from .translation import translation
_ = translation.gettext_lazy

from .utils import get_credit_card_issuer

class DateValidator(object):
    def __init__(self, message=None):
        if not message:
            message = _('Please enter a valid date.')
        self.message = message
    def __call__(self, form, field):
        data = field.data
        if isinstance(data, str):
            data = datetime.datetime.strptime(data, '%Y-%m').date()
        if data < datetime.date.today():
            raise ValidationError(self.message)

class CreditCardNumberValidator(object):
    def __init__(self, message=None):
        if not message:
            message = _('Please enter a valid card number')
        self.message = message

    def __call__(self, form, field):
        if not self.cart_number_checksum_validation(field.data):
            raise ValidationError(self.message)

    @staticmethod
    def cart_number_checksum_validation(number):
        digits = []
        even = False
        if not number.isdigit():
            return False
        for digit in reversed(number):
            digit = ord(digit) - ord('0')
            if even:
                digit *= 2
                if digit >= 10:
                    digit = digit % 10 + digit // 10
            digits.append(digit)
            even = not even
        return sum(digits) % 10 == 0 if digits else False

class PaymentForm(Form):
    '''
    Payment form

    When displaying the form remember to use *action* and *method*.
    '''
    method = 'post'
    action = ''
    provider = None
    payment = None

    def __init__(self, *, provider=None, payment=None, **kwargs):
        if "data" not in kwargs:
            kwargs["obj"] = payment
        super().__init__(**kwargs)
        self.provider = provider
        self.payment = payment
        if provider and payment:
            self.action = provider.get_action(payment)

class CreditCardPaymentForm(PaymentForm):
    # which credit card types are accepted?
    VALID_TYPES = None

    number = StringField(label=_('Card Number'),
        validators=[validators.Length(max=32),
                    validators.Required(), CreditCardNumberValidator()],
        render_kw={'autocomplete': 'cc-number'})

    expiration = DateField(label=_('Expiration date (YYYY-MM):'),
        validators=[DateValidator()],
        format='%Y-%m',
        render_kw={'autocomplete': 'cc-exp'})

    cvv2 = StringField(
        label=_('CVV2 Security Number'), validators=[validators.Required(_('Enter a valid security number.')), validators.Regexp('^[0-9]{3,4}$', message=_('Enter a valid security number.'))],
        description=_(
            'Last three digits located on the back of your card.'
            ' For American Express the four digits found on the front side.'),
        render_kw={'autocomplete': 'cc-csc'})

    def __init__(self, *, valid_types=None, **kwargs):
        self.VALID_TYPES = valid_types
        super().__init__(**kwargs)

    def validate_number(self, field):
        if self.VALID_TYPES and get_credit_card_issuer(field.data)[0] not in self.VALID_TYPES:
            raise ValidationError(
                _('We accept only %(valid_types)s') % {"valid_types": ", ".join(self.VALID_TYPES)})


class CreditCardPaymentFormWithName(CreditCardPaymentForm):
    name = StringField(label=_('Name on Credit Card'),
        validators=[validators.Length(max=128)])




import json
from uuid import uuid4
from decimal import Decimal

from django.conf import settings
from django.db import models
from ..translation import translation
_ = translation.gettext_lazy


try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse

from .. import FraudStatus, PaymentStatus
from ..logic import BasePaymentLogic
from .signals import status_changed
from .utils import add_prefixed_address
from ..utils import getter_prefixed_address

class BasePayment(models.Model, BasePaymentLogic):
    '''
    Represents a single transaction. Each instance has one or more PaymentItem.
    '''
    variant = models.CharField(max_length=255)
    #: Transaction status
    status = models.CharField(
        max_length=10, choices=PaymentStatus.CHOICES,
        default=PaymentStatus.WAITING)
    fraud_status = models.CharField(
        max_length=10, choices=FraudStatus.CHOICES,
        default=FraudStatus.UNKNOWN)
    fraud_message = models.TextField(blank=True, default='')
    #: Creation date and time
    created = models.DateTimeField(auto_now_add=True)
    #: Date and time of last modification
    modified = models.DateTimeField(auto_now=True)
    #: Transaction ID (if applicable)
    transaction_id = models.CharField(max_length=255, blank=True)
    #: Currency code (may be provider-specific)
    currency = models.CharField(max_length=10)

    description = models.TextField(blank=True, default='')
    billing_email = models.EmailField(blank=True)
    customer_ip_address = models.GenericIPAddressField(blank=True, null=True)
    extra_data = models.TextField(blank=True, default='')
    message = models.TextField(blank=True, default='')
    token = models.CharField(max_length=36, blank=True, default='')

    #: Total amount (gross)
    total = models.DecimalField(max_digits=20, decimal_places=8, default=Decimal('0.0'))
    #: Delivery costs
    delivery = models.DecimalField(
        max_digits=20, decimal_places=8, default=Decimal('0.0'))
    #: Tax
    tax = models.DecimalField(max_digits=20, decimal_places=8, default=Decimal('0.0'))
    #: captured = current captured amount
    captured_amount = models.DecimalField(
        max_digits=20, decimal_places=8, default=Decimal('0.0'))

    class Meta:
        abstract = True


    def signal_status_change(self):
        status_changed.send(sender=type(self), instance=self)

    @classmethod
    def check_token_exists(cls, token):
        return cls._default_manager.filter(token=token).exists()

    def get_process_url(self):
        return reverse('process_payment', kwargs={'token': self.token})

    def save(self, **kwargs):
        self.create_token()
        return models.Model.save(self, **kwargs)

@add_prefixed_address("billing")
class BasePaymentWithAddress(BasePayment):
    """ Has real billing address + shippingaddress alias on billing address (alias for backward compatibility) """
    get_billing_address = getter_prefixed_address("billing")
    get_shipping_address = get_billing_address

    class Meta:
        abstract = True



import os
import gettext
from functools import partial, lru_cache
from .core import get_language

web_payments_translation_path = os.path.join(os.path.dirname(__file__), "locale")


class Translation(object):
    _fallback = None
    domain = None
    translation_path = None
    def __init__(self, translation_path, domain="web_payments", fallback=None):
        self.translation_path = translation_path
        self.domain = domain
        if fallback:
            self._fallback = fallback
        else:
            self._fallback = []

    @lru_cache(typed=True)
    def _trans(self, lang):
        return gettext.translation("web_payments", self.translation_path, [lang]+self._fallback, fallback=True)

    def trans(self, cur_lang=None):
        if not cur_lang:
            cur_lang = get_language()
        return self._trans(cur_lang)

    def gettext(self, msg, cur_lang=None):
        return self.trans(cur_lang).gettext(msg)

    def ngettext(self, msg, msgplural, n, cur_lang=None):
        return self.trans(cur_lang).ngettext(msg, msgplural, n)

    def gettext_lazy(self, msg, cur_lang=None):
        return partial(self.gettext, msg, cur_lang)

    def ngettext_lazy(self, msg, msgplural, n, cur_lang=None):
        return partial(self.ngettext, msg, msgplural, n, cur_lang)

translation = Translation(web_payments_translation_path, domain="web_payments")




from urllib.error import URLError
from urllib.parse import urlencode
from unittest import TestCase
from unittest.mock import MagicMock

from . import DummyProvider
from web_payments import PaymentError, RedirectNeeded
from web_payments.status import FraudStatus, PaymentStatus
from web_payments.testcommon import create_test_payment

VARIANT = 'dummy-3ds'


Payment = create_test_payment(variant=VARIANT)


class TestDummy3DSProvider(TestCase):

    def setUp(self):
        self.payment = Payment()

    def test_process_data_supports_verification_result(self):
        provider = DummyProvider()
        verification_status = PaymentStatus.CONFIRMED
        request = MagicMock()
        request.GET = {'verification_result': verification_status}
        with self.assertRaises(RedirectNeeded) as exc:
            provider.process_data(self.payment, request)
            self.assertEqual(self.payment.status, verification_status)
            self.assertEqual(exc.args[0], self.payment.get_success_url())

    def test_process_data_redirects_to_success_on_payment_success(self):
        self.payment.status = PaymentStatus.PREAUTH
        provider = DummyProvider()
        request = MagicMock()
        request.GET = {}
        with self.assertRaises(RedirectNeeded) as exc:
            provider.process_data(self.payment, request)
            self.assertEqual(exc.args[0], self.payment.get_success_url())

    def test_process_data_redirects_to_failure_on_payment_failure(self):
        self.payment.status = PaymentStatus.REJECTED
        provider = DummyProvider()
        request = MagicMock()
        request.GET = {}
        with self.assertRaises(RedirectNeeded) as exc:
            provider.process_data(self.payment, request)
            self.assertEqual(exc.args[0], self.payment.get_failure_url())

    def test_provider_supports_non_3ds_transactions(self):
        provider = DummyProvider()
        data = {
            'status': PaymentStatus.PREAUTH,
            'fraud_status': FraudStatus.UNKNOWN,
            'gateway_response': '3ds-disabled',
            'verification_result': ''
        }
        with self.assertRaises(RedirectNeeded) as exc:
            provider.get_form(self.payment, data)
            self.assertEqual(exc.args[0], self.payment.get_success_url())

    def test_provider_raises_verification_result_needed_on_success(self):
        provider = DummyProvider()
        data = {
            'status': PaymentStatus.WAITING,
            'fraud_status': FraudStatus.UNKNOWN,
            'gateway_response': '3ds-redirect'
        }

        form = provider.get_form(self.payment, data)
        self.assertFalse(form.validate())

    def test_provider_supports_3ds_redirect(self):
        provider = DummyProvider()
        verification_result = PaymentStatus.CONFIRMED
        data = {
            'status': PaymentStatus.WAITING,
            'fraud_status': FraudStatus.UNKNOWN,
            'gateway_response': '3ds-redirect',
            'verification_result': verification_result
        }
        params = urlencode({'verification_result': verification_result})
        expected_redirect = '%s?%s' % (self.payment.get_process_url(), params)

        with self.assertRaises(RedirectNeeded) as exc:
            provider.get_form(self.payment, data)
            self.assertEqual(exc.args[0], expected_redirect)

    def test_provider_supports_gateway_failure(self):
        provider = DummyProvider()
        data = {
            'status': PaymentStatus.WAITING,
            'fraud_status': FraudStatus.UNKNOWN,
            'gateway_response': 'failure',
            'verification_result': ''
        }

        with self.assertRaises(URLError):
            provider.get_form(self.payment, data)

    def test_provider_raises_redirect_needed_on_success(self):
        provider = DummyProvider()
        data = {
            'status': PaymentStatus.PREAUTH,
            'fraud_status': FraudStatus.UNKNOWN,
            'gateway_response': '3ds-disabled',
            'verification_result': ''
        }

        with self.assertRaises(RedirectNeeded) as exc:
            provider.get_form(self.payment, data)
            self.assertEqual(exc.args[0], self.payment.get_success_url())

    def test_provider_raises_redirect_needed_on_failure(self):
        provider = DummyProvider()
        data = {
            'status': PaymentStatus.ERROR,
            'fraud_status': FraudStatus.UNKNOWN,
            'gateway_response': '3ds-disabled',
            'verification_result': ''
        }

        with self.assertRaises(RedirectNeeded) as exc:
            provider.get_form(self.payment, data)
            self.assertEqual(exc.args[0], self.payment.get_failure_url())

    def test_provider_raises_payment_error(self):
        provider = DummyProvider()
        data = {
            'status': PaymentStatus.PREAUTH,
            'fraud_status': FraudStatus.UNKNOWN,
            'gateway_response': 'payment-error',
            'verification_result': ''
        }

        with self.assertRaises(PaymentError):
            provider.get_form(self.payment, data)

    def test_provider_switches_payment_status_on_get_form(self):
        provider = DummyProvider()
        form = provider.get_form(self.payment, data={})
        self.assertEqual(self.payment.status, PaymentStatus.INPUT)




from urllib.error import URLError
from urllib.parse import urlencode
from django.http import HttpResponseRedirect

from .forms import DummyForm
from web_payments import PaymentError, RedirectNeeded
from web_payments.status import PaymentStatus
from web_payments.logic import BasicProvider

class DummyProvider(BasicProvider):
    '''
    Dummy payment provider
    '''

    def get_form(self, payment, data=None):
        if payment.status == PaymentStatus.WAITING:
            payment.change_status(PaymentStatus.INPUT)
        form = DummyForm(data=data, provider=self,
                         payment=payment)
        if form.validate():
            new_status = form.status.data
            payment.change_status(new_status)
            new_fraud_status = form.fraud_status.data
            payment.change_fraud_status(new_fraud_status)

            gateway_response = form.gateway_response.data
            verification_result = form.verification_result.data
            if gateway_response or verification_result:
                if gateway_response == '3ds-disabled':
                    # Standard request without 3DSecure
                    pass
                elif gateway_response == '3ds-redirect':
                    # Simulate redirect to 3DS and get back to normal
                    # payment processing
                    process_url = payment.get_process_url()
                    params = urlencode(
                        {'verification_result': verification_result})
                    redirect_url = '%s?%s' % (process_url, params)
                    raise RedirectNeeded(redirect_url)
                elif gateway_response == 'failure':
                    # Gateway raises error (HTTP 500 for example)
                    raise URLError('Opps')
                elif gateway_response == 'payment-error':
                    raise PaymentError('Unsupported operation')

            if new_status in [PaymentStatus.PREAUTH, PaymentStatus.CONFIRMED]:
                raise RedirectNeeded(payment.get_success_url())
            raise RedirectNeeded(payment.get_failure_url())
        return form

    def process_data(self, payment, request):
        verification_result = request.GET.get('verification_result')
        if verification_result:
            payment.change_status(verification_result)
        if payment.status in [PaymentStatus.CONFIRMED, PaymentStatus.PREAUTH]:
            raise RedirectNeeded(payment.get_success_url())
        raise RedirectNeeded(payment.get_failure_url())

    def capture(self, payment, amount=None, final=True):
        payment.change_status(PaymentStatus.CONFIRMED)
        return amount

    def release(self, payment):
        return None

    def refund(self, payment, amount=None):
        return amount or 0



from django.db import models

def add_prefixed_address(prefix):
    """ add address with prefix to class """
    first_name = "{}_first_name".format(prefix)
    last_name = "{}_last_name".format(prefix)
    address_1 = "{}_address_1".format(prefix)
    address_2 = "{}_address_2".format(prefix)
    city = "{}_city".format(prefix)
    postcode = "{}_postcode".format(prefix)
    country_code = "{}_country_code".format(prefix)
    country_area = "{}_country_area".format(prefix)
    def class_to_customize(dclass):
        setattr(dclass, first_name, models.CharField(max_length=256, blank=True))
        setattr(dclass, last_name, models.CharField(max_length=256, blank=True))
        setattr(dclass, address_1, models.CharField(max_length=256, blank=True))
        setattr(dclass, address_2, models.CharField(max_length=256, blank=True))
        setattr(dclass, city, models.CharField(max_length=256, blank=True))
        setattr(dclass, postcode, models.CharField(max_length=256, blank=True))
        setattr(dclass, country_code, models.CharField(max_length=2, blank=True))
        setattr(dclass, country_area, models.CharField(max_length=256, blank=True))
        return dclass
    return class_to_customize



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
# apps clash with 'apps' of module so import as django_apps
from django.apps import apps as django_apps
from django.utils.translation import get_language

from .. import core

PAYMENT_HOST = None
PAYMENT_USES_SSL = None

default_app_config = 'web_payments.django.apps.WebPaymentsConfig'

def get_payment_model():
    '''
    Return the Payment model that is active in this project
    '''
    try:
        app_label, model_name = settings.PAYMENT_MODEL.split('.')
    except (ValueError, AttributeError):
        raise ImproperlyConfigured('PAYMENT_MODEL must be of the form '
                                   '"app_label.model_name"')
    payment_model = django_apps.get_model(app_label, model_name)
    if payment_model is None:
        msg = (
            'PAYMENT_MODEL refers to model "%s" that has not been installed' %
            settings.PAYMENT_MODEL)
        raise ImproperlyConfigured(msg)
    return payment_model

def get_base_url(variant=None):
    """
    Returns host url according to project settings. Protocol is chosen by
    checking PAYMENT_USES_SSL variable.
    If PAYMENT_HOST is not specified, gets domain from Sites.
    Otherwise checks if it's callable and returns it's result. If it's not a
    callable treats it as domain.
    """
    from django.contrib.sites.models import Site
    protocol = 'https' if PAYMENT_USES_SSL else 'http'
    if not PAYMENT_HOST:
        current_site = Site.objects.get_current()
        domain = current_site.domain
    elif callable(PAYMENT_HOST):
        domain = PAYMENT_HOST()
    else:
        domain = PAYMENT_HOST
    return '%s://%s' % (protocol, domain)



def load_settings(initialize=None):
    ''' loads settings and sets functions, required for initialization
        default: initialize only if not initialized
     '''
    global PAYMENT_HOST
    global PAYMENT_USES_SSL

    if initialize is None:
        initialize = not core.is_initialized

    if getattr(settings, "PAYMENT_VARIANTS_API", None):
        core.PAYMENT_VARIANTS_API = settings.PAYMENT_VARIANTS_API


    PAYMENT_HOST = getattr(settings, 'PAYMENT_HOST', None)
    if not PAYMENT_HOST:
        if 'django.contrib.sites' not in settings.INSTALLED_APPS:
            raise ImproperlyConfigured('The PAYMENT_HOST setting without '
                                       'the sites app must not be empty.')
        from django.contrib.sites.models import Site

    PAYMENT_USES_SSL = getattr(settings, 'PAYMENT_USES_SSL', not settings.DEBUG)


    if initialize:
        core.get_payment_model = get_payment_model
        core.get_base_url = get_base_url
        core.get_language = get_language
        core.is_initialized = True
